////////////////////////////////////////////////////////////////////////////////
//
//   LLVMTA - Timing Analyser performing worst-case execution time analysis
//     using the LLVM backend intermediate representation
//
// Copyright (C) 2013-2015  Saarland University
// Copyright (C) 2014-2015  Claus Faymonville
//
// This file is distributed under the Saarland University Software Release
// License. See LICENSE.TXT for details.
//
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING BUT NOT LIMITED TO, THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE SAARLAND UNIVERSITY, THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING BUT NOT LIMITED TO PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED OR OTHER LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, TORT OR OTHERWISE, ARISING IN ANY WAY FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH A DAMAGE.
//
////////////////////////////////////////////////////////////////////////////////

#ifndef TARGET_ARM_TIMINGANALYSISMAIN_H
#define TARGET_ARM_TIMINGANALYSISMAIN_H

#include "ARM.h"
#include "ARMMachineFunctionInfo.h"
#include "llvm/CodeGen/MachineFunctionPass.h"
#include "llvm/CodeGen/MachineModuleInfo.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/IR/Module.h"
#include "ARMAsmPrinter.h"

#include "AnalysisFramework/AnalysisResults.h"
#include "Util/Options.h"

#include <iostream>
#include <map>

using namespace llvm;

namespace TimingAnalysisPass {

	// Forward declaration
	class AddressInformation;

  /**
   * This is the main pass of the timing analysis module.
   * At the end of all Machine-Level passes except CodeEmitter, the timing analysis is triggered.
   * The analysis entry point is given via commandline option.
   * This pass triggers value, microarchitectural, and path analyses.
   * The output is an ILP formulation whose solution is a WCET bound.
   */
  class TimingAnalysisMain : public MachineFunctionPass {
  
  public:
    static char ID;
    TimingAnalysisMain(TargetMachine& TM);
		
		/**
		 * This is a dummy function.
		 */
    bool runOnMachineFunction(MachineFunction &F);
		
		/**
		 * The timing analysis starts when all previous passes are finished on ALL functions.
		 * The timing analysis, i.e. value, microarchitectural, and path analyses are triggered here.
		 */
		virtual bool doFinalization(Module& M);

		virtual llvm::StringRef getPassName() const {
      return "TA: Main phase of Timing Analysis, e.g. Value and Microarchitectural Analyses";
    }

		/**
		 * Provide static access to detailed information about the target machine.
		 */
		static TargetMachine& getTargetMachine();

  private:
		
		/** 
		 * Dispatch the value analysis which is specific to the used ISA.
		 */
		template <Triple::ArchType ISA>
  	void dispatchValueAnalysis();

		/**
		 * Select the overall analysis type, so timing, memory access, and cache
		 */
		void dispatchAnalysisType(AddressInformation& addressInfo);

		///////////////////////////
		// Timing/Cache Analysis //
		///////////////////////////

		// Dispatch the individual microarchitectures with different memory topologies for timing analysis
		boost::optional<BoundItv> dispatchTimingAnalysis(AddressInformation& addressInfo);
		boost::optional<BoundItv> dispatchCacheAnalysis(AnalysisType anatype, AddressInformation& addressInfo);

		// Private fiels
  	static TargetMachine *targetMachine;
  };

	/**
	 * Provide access to important options. Here: The initial stack pointer.
	 */
  unsigned getInitialStackPointer();
  /**
   * Provide access to important options. Here: The initial link register.
   */
  unsigned getInitialLinkRegister();
  /**
   * Provide access to important options. Here: The analysis entry point.
   */
  MachineFunction* getAnalysisEntryPoint();
}

namespace llvm {
	FunctionPass *createTimingAnalysisMain(TargetMachine &TM);
}

#endif
