////////////////////////////////////////////////////////////////////////////////
//
//   LLVMTA - Timing Analyser performing worst-case execution time analysis
//     using the LLVM backend intermediate representation
//
// Copyright (C) 2013-2015  Saarland University
//
// This file is distributed under the Saarland University Software Release
// License. See LICENSE.TXT for details.
//
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING BUT NOT LIMITED TO, THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE SAARLAND UNIVERSITY, THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING BUT NOT LIMITED TO PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED OR OTHER LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, TORT OR OTHERWISE, ARISING IN ANY WAY FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH A DAMAGE.
//
////////////////////////////////////////////////////////////////////////////////

#ifndef DIRECTIVEHEURISTICS_H
#define DIRECTIVEHEURISTICS_H

#include "ARM.h"
#include "ARMMachineFunctionInfo.h"
#include "llvm/CodeGen/MachineFunctionPass.h"
#include "llvm/CodeGen/MachineLoopInfo.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/CodeGen/TargetInstrInfo.h"
#include <iostream>
#include <map>
#include <list>

#include "PartitionUtil/Directive.h"
#include "Util/Util.h"

using namespace llvm;

namespace TimingAnalysisPass {

	/**
	 * Class that is responsible for deriving partitioning directives.
	 * It is implemented as a MachineFunctionPass within the LLVM Framework.
	 * It derives directives on the basic block level, as well as on the instruction level.
	 */
	class DirectiveHeuristicsPass : public MachineFunctionPass {
		public:
		static char ID;
		DirectiveHeuristicsPass();

		~DirectiveHeuristicsPass() {
			// TODO get rid of all these pointers in the map
			// delete all directives
			for (auto ele : directiveBeforeInstr) {
				delete ele.second;
			}
			for (auto ele : directiveAfterInstr) {
				delete ele.second;
			}
			for (auto ele : directiveEnterBBEdge) {
				for (auto ele2 : *ele.second) {
					delete ele2;
				}
				delete ele.second;
			}
			for (auto ele : directiveLeaveBBEdge) {
				for (auto ele2 : *ele.second) {
					delete ele2;
				}
				delete ele.second;
			}
			for (auto ele : directiveBeforeFunc) {
				delete ele.second;
			}
			for (auto ele : directiveAfterFunc) {
				delete ele.second;
			}
		}

		/**
		 * Declare dependencies of this analysis, here the MachineLoopInfo.
		 */
		void getAnalysisUsage(llvm::AnalysisUsage &AU) const;

		/**
		 * Derive partitioning directives on the basic block level, for each function.
		 * This include the following directives currently:
		 * - Directives for Partitions of Type ::FUNCALLEE on function entry and exit
		 * - Directives for Partitions of Type ::LOOPPEEL and ::LOOPITER on entry and exit
		 *   of looping regions
		 */
		bool runOnMachineFunction(MachineFunction &MF);
		/**
		 * Derive partitioning directives on the instruction level, for each basic block.
		 * This includes directives of type ::CREATE/::MERGE for Partitions ::CALLSITE on each
		 * function call site.
		 */
		bool runOnMachineBasicBlock(MachineBasicBlock &MBB);

		llvm::StringRef getPassName() const {
			return "Annotate the control-flow-graph with Trace Partitioning Directives";
		}
		
		/**
		 * Dump all annotated directives. For debugging purposes.
		 */
		void dump(std::ostream& mystream) const;
		
		/**
		 * Returns the directive that should be executed directly before the given instruction MI.
		 */
		Directive* getDirectiveBeforeInstr(const MachineInstr* MI);
		/**
		 * Returns the directive that should be executed directly after the given instruction MI.
		 */
		Directive* getDirectiveAfterInstr(const MachineInstr* MI);
		/**
		 * Returns the directive that should be executed as first thing on call of the given function MF.
		 */
		Directive* getDirectiveOnCall(const MachineFunction* MF);
		/**
		 * Returns the directive that should be executed as last thing on return of the given function MF.
		 */
		Directive* getDirectiveOnReturn(const MachineFunction* MF);
		/**
		 * Returns directives that should be executed directly when the given edge is entered.
		 * The result implies an ordering: First list members are to be executed first.
		 */
		std::list<Directive*>* getDirectiveOnEdgeEnter(MBBedge E);
		/**
		 * Returns directives that should be executed directly before leaving the given edge.
		 * The result implies an ordering: First list members are to be executed first.
		 */
		std::list<Directive*>* getDirectiveOnEdgeLeave(MBBedge E);

		/**
		 * Checks for the directive that should be executed directly before the given instruction MI.
		 */
		bool hasDirectiveBeforeInstr(const MachineInstr* MI);
		/**
		 * Checks for the directive that should be executed directly after the given instruction MI.
		 */
		bool hasDirectiveAfterInstr(const MachineInstr* MI);
		/**
		 * Checks for the directive that should be executed as first thing on call of the given function.
		 */
		bool hasDirectiveOnCall(const MachineFunction* MF);
		/**
		 * Checks for the directive that should be executed as last thing on return of the given function.
		 */
		bool hasDirectiveOnReturn(const MachineFunction* MF);
		/**
		 * Checks for directives that should be executed directly when the given edge is entered.
		 */
		bool hasDirectiveOnEdgeEnter(MBBedge E);
		/**
		 * Checks for directives that should be executed directly before leaving the given edge.
		 */
		bool hasDirectiveOnEdgeLeave(MBBedge E);

		private:
		///< Directive to be executed before the given instruction
		std::map<const MachineInstr*, Directive*, instrptrcomp> directiveBeforeInstr;
		///< Directive to be executed after the given instruction
		std::map<const MachineInstr*, Directive*, instrptrcomp> directiveAfterInstr;

		///< Directive to be executed when the given edge between basicblocks is entered.
		///< Directives are executed in list-order from begin to end.
		std::map<MBBedge, std::list<Directive*>*, mbbedgecomp> directiveEnterBBEdge;
		///< Directive to be executed when the given edge between basicblocks is left.
		///< Directives are executed in list-order from begin to end.
		std::map<MBBedge, std::list<Directive*>*, mbbedgecomp> directiveLeaveBBEdge;

		///< Directive to be executed before the given function
		std::map<const MachineFunction*, Directive*, machfunccomp> directiveBeforeFunc;
		///< Directive to be executed after the given function
		std::map<const MachineFunction*, Directive*, machfunccomp> directiveAfterFunc;

		/**
		 * Annotate partitioning directives for loop peeling for all loops 
		 * in the current (called from runOnMachineFunction) function.
		 * The heuristics peels each loop as many times as is given by the -ta-loop-peel option.
		 */
		void annotateLoopDirective(const llvm::MachineLoop* loop);
	};

	/**
	 * Singleton Object of the Directive Derivation Pass.
	 * Used to access the derived directives later on.
	 */
	extern DirectiveHeuristicsPass* directiveHeuristicsPass;

}

namespace llvm {
	FunctionPass* createDirectiveHeuristicsPass();
}

#endif
