////////////////////////////////////////////////////////////////////////////////
//
//   LLVMTA - Timing Analyser performing worst-case execution time analysis
//     using the LLVM backend intermediate representation
//
// Copyright (C) 2013-2015  Saarland University
// Copyright (C) 2014-2015  Claus Faymonville
//
// This file is distributed under the Saarland University Software Release
// License. See LICENSE.TXT for details.
//
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING BUT NOT LIMITED TO, THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE SAARLAND UNIVERSITY, THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING BUT NOT LIMITED TO PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED OR OTHER LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, TORT OR OTHERWISE, ARISING IN ANY WAY FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH A DAMAGE.
//
////////////////////////////////////////////////////////////////////////////////

#ifndef LOOPBOUNDINFO_H
#define LOOPBOUNDINFO_H

#include "ARM.h"
#include "llvm/CodeGen/MachineFunctionPass.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/CodeGen/MachineLoopInfo.h"
#include "llvm/Analysis/ScalarEvolution.h"

#include "PreprocessingAnalysis/ConstantValueDomain.h"
#include "AnalysisFramework/PartitioningDomain.h"
#include "PartitionUtil/Context.h"

#include <iostream>
#include <map>
#include <list>

namespace TimingAnalysisPass {
 
 	/**
 	 * Machine function pass that extracts information about potential loop bounds from the
 	 * LLVM middle-end, i.e. from LoopInfo and ScalarEvolution, respectively.
 	 */
  class LoopBoundInfoPass : public llvm::MachineFunctionPass {
  
  public:
    static char ID;
    LoopBoundInfoPass();
	
		/**
		 * This is a dummy function.
		 */
    bool runOnMachineFunction(llvm::MachineFunction &MF);
		
		void getAnalysisUsage(llvm::AnalysisUsage &AU) const;

		virtual llvm::StringRef getPassName() const {
      return "Provides loop bound information extracted from ScalarEvolution and LoopInfo";
    }

		template <llvm::Triple::ArchType ISA>
		using CVAnaDom = PartitioningDomain<ConstantValueDomain<ISA>, llvm::MachineInstr>;

		template <llvm::Triple::ArchType ISA>
		using CVAnalysisType = AnalysisInformation<CVAnaDom<ISA>, llvm::MachineInstr>;

		/**
	 	 * Dump the computed instruction<->address mapping to the given stream.
	 	 */
		void dump(std::ostream& mystream) const;
		
		/**
		 * Get all machine loops. This are valid as we modified LLVM not to throw them away.
		 */
		const std::list<const llvm::MachineLoop*> getAllLoops() const;

		/**
		 * Is there a known bound for this loop?
		 */
		bool hasUpperLoopBound(const llvm::MachineLoop* loop, const Context& ctx) const;
		bool hasLowerLoopBound(const llvm::MachineLoop* loop, const Context& ctx) const;

		/**
		 * Compute SCEVs to derive the actual loopbounds using the information
		 * from the Constant Value Analysis.
		 */
		template <llvm::Triple::ArchType ISA>
		void computeLoopBoundFromCVDomain(const CVAnalysisType<ISA>& cvAnaInfo);

		/**
		 * Get the loop bound if any.
		 * The bound is the number of times the backedge can be taken per entrance of the loop.
		 * Note this is one less than the loop iteration count per entrance.
		 */
		unsigned getUpperLoopBound(const llvm::MachineLoop* loop, const Context& ctx) const;
		unsigned getLowerLoopBound(const llvm::MachineLoop* loop, const Context& ctx) const;

		std::string getLoopDesc(const llvm::MachineLoop* loop) const;

		void dumpNonUpperBoundLoops(std::ostream& mystream, std::ostream& mystream2) const;
		void dumpNonLowerBoundLoops(std::ostream& mystream, std::ostream& mystream2) const;

		void parseManualUpperLoopBounds(const char* filename);
		void parseManualLowerLoopBounds(const char* filename);
		
  private:
		/**
		 * Private helpers for four pairs of public member functions.
		 * Introduced to avoid code duplication and to guarantee maintainability.
		 */

		bool hasLoopBoundNoCtx(const llvm::MachineLoop* loop,
				const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& loopBounds,
				const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& manualLoopBounds,
				const std::unordered_map<const llvm::MachineLoop*, unsigned>& manualLoopBoundsNoCtx
				) const;
		bool hasLoopBound(const llvm::MachineLoop* loop,
				const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& loopBounds,
				const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& manualLoopBounds,
				const std::unordered_map<const llvm::MachineLoop*, unsigned>& manualLoopBoundsNoCtx,
				const Context& ctx
				) const;

		unsigned getLoopBoundNoCtx(const llvm::MachineLoop* loop,
				const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& loopBounds,
				const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& manualLoopBounds,
				const std::unordered_map<const llvm::MachineLoop*, unsigned>& manualLoopBoundsNoCtx
				) const;

		unsigned getLoopBound(const llvm::MachineLoop* loop,
				const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& loopBounds,
				const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& manualLoopBounds,
				const std::unordered_map<const llvm::MachineLoop*, unsigned>& manualLoopBoundsNoCtx,
				const Context& ctx
				) const;
		void dumpNonBoundLoops(std::ostream& mystream,
				std::ostream& mystream2,
				bool (LoopBoundInfoPass::* hasLoopBoundFkt) (const llvm::MachineLoop* loop, const Context& ctx) const,
				unsigned (LoopBoundInfoPass::* getLoopBoundFkt) (const llvm::MachineLoop* loop, const Context& ctx) const
				) const;
		void parseManualLoopBounds(const char* filename,
				std::unordered_map<const llvm::MachineLoop*,
				std::unordered_map<Context, unsigned>>& manualLoopBounds,
				std::unordered_map<const llvm::MachineLoop*, unsigned>& manualNormalLoopBounds
				);

		/**
		 * Internal function: Walks over all recognized loops and searches for bounds using
		 * ScalarEvolution.
		 */
		void walkLoop(const llvm::Loop* loop);
		/**
		 * Internal function: Walks over all recognized machine-loops, matches them with 
		 * middle-end-IR-loops, and stores the loop bounds for later use.
		 */
		void walkMachineLoop(const llvm::MachineLoop* loop);
		/**
		 * Helper function to determine whether each BB in maloop and irloop has a match.
		 */
		bool isMachineLoopExactMatch(const llvm::MachineLoop* maloop, const llvm::Loop* irloop);
		/**
		 * Helper function to determine whether each BB in maloop has a match.
		 */
		bool isMachineLoopPartialMatch(const llvm::MachineLoop* maloop, const llvm::Loop* irloop);
		/**
		 * Internal Function: Map the function to the SCEV that represents its
		 * loopbound.
		 */
		void addSCEVMapping(const llvm::MachineLoop* ML, const llvm::Loop *loop);

		/**
		 * Internal Function: Compute the loopbound for all loops from all
		 * callsites using their SCEVs and the Constant Value Domain and
		 * store them in a map.
		 */
		template <llvm::Triple::ArchType ISA>
		void computeLoopBounds(const std::unordered_map<const llvm::MachineLoop*, const llvm::SCEV*>& loopMapping,
				std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& loopBounds,
				const CVAnalysisType<ISA>& cvAnaInfo);

		/**
		 * Get the Context Sensitive Loop Bounds.
		 * Given a SCEV and the Constant Value Domain Analysis Information,
		 * evaluate the SCEV for each context in AnaInfo and return a map with
		 * loopbounds per Context.
		 */
		template <llvm::Triple::ArchType ISA>
		std::unordered_map<Context, unsigned> getContextSensitiveBounds(const llvm::MachineLoop* loop, const llvm::SCEV* equation,
				const std::unordered_map<Context, ConstantValueDomain<ISA>>& anaInfo);

		/**
		 * Compute the value of a SCEV using the CV Domain to get unknown
		 * values.
		 */
		template <llvm::Triple::ArchType ISA>
		bool getSCEVBoundFromCVDomain(const SCEV *equation, const ConstantValueDomain<ISA>& CVDom, unsigned& value);

		/**
		 * Get a value from the CV Domain for a function argument.
		 *
		 * Given an argument number and a Constant Value Domain, identify the
		 * function argument and get its value from the CV Domain.
		 */
		template <llvm::Triple::ArchType ISA>
		bool getValueByArgument(const llvm::Argument* arg, const ConstantValueDomain<ISA>& CVDom, unsigned& value);

		void parseCtxSensManualLoopBounds(std::ifstream& file,
				std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& manualLoopBounds
				);
		void parseNormalManualLoopBounds(std::ifstream& file,
				std::unordered_map<const llvm::MachineLoop*, unsigned>& manualLoopBounds
				);

		// USING LOOPS AND MACHINELOOPS SHOULD BE SAFE AS WE MODIFIED LLVM ACCORDINGLY.
		/**
		 * The set of recognized middle-end IR loops.
		 */
		std::list<const llvm::Loop*> irLoops;
		/**
		 * The set of recognized machine loops.
		 */
		std::list<const llvm::MachineLoop*> maLoops;
		/**
		 * A computed mapping between machine and middle-end ir loops.
		 *
		 * This is currently only required for dumping the unknown loopbounds.
		 * Sometimes, compiler optimizations cause some information to be lost
		 * and we need this mapping to find the required symbols.
		 */
		std::map<const llvm::MachineLoop*, const llvm::Loop*> loopMapping;

		/**
		 * Loop bounds for middle-end ir loops as found by ScalarEvolution.
		 * Stores the max-backedge-taken-count that is one less than the maximal execution count
		 * of the loop header.
		 */
		std::unordered_map<const llvm::MachineLoop*, std::unordered_set<Context>> loopContextMap;
		std::unordered_map<const llvm::MachineLoop*, const llvm::SCEV*> upperLoopBoundsSCEV;
		std::unordered_map<const llvm::MachineLoop*, const llvm::SCEV*> lowerLoopBoundsSCEV;
		std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>> upperLoopBoundsCtx;
		std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>> lowerLoopBoundsCtx;

		/**
		 * Manual loop bounds (backedge taken counts per loop entrance) annotated by the user of llvmta.
		 * This is a fallback for unfound loops from the ir-representation.
		 */
		std::unordered_map<const llvm::MachineLoop*, unsigned> manualUpperLoopBoundsNoCtx;
		std::unordered_map<const llvm::MachineLoop*, unsigned> manualLowerLoopBoundsNoCtx;
		std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>> manualUpperLoopBounds;
		std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>> manualLowerLoopBounds;
  };

	extern LoopBoundInfoPass* loopBoundInfoPass;

}

namespace llvm {
	FunctionPass *createLoopBoundInfoPass();
}

#endif
