////////////////////////////////////////////////////////////////////////////////
//
//   LLVMTA - Timing Analyser performing worst-case execution time analysis
//     using the LLVM backend intermediate representation
//
// Copyright (C) 2013-2015  Saarland University
// Copyright (C) 2014-2015  Claus Faymonville
//
// This file is distributed under the Saarland University Software Release
// License. See LICENSE.TXT for details.
//
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING BUT NOT LIMITED TO, THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE SAARLAND UNIVERSITY, THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING BUT NOT LIMITED TO PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED OR OTHER LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, TORT OR OTHERWISE, ARISING IN ANY WAY FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH A DAMAGE.
//
////////////////////////////////////////////////////////////////////////////////

#include "LLVMPasses/StaticAddressProvider.h"

#include "RISCV.h"

#include "Util/Util.h"
#include "Util/Options.h"
#include "AnalysisFramework/CallGraph.h"

#include "llvm/CodeGen/MachineConstantPool.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/raw_os_ostream.h"
#include "llvm/CodeGen/MachineJumpTableInfo.h"
#include "llvm/IR/Constants.h"

#include <sstream>

using namespace llvm;

namespace TimingAnalysisPass {

StaticAddressProvider* staticAddressProvider;

char StaticAddressProvider::ID = 0;

StaticAddressProvider::StaticAddressProvider(TargetMachine &TM)
			: MachineFunctionPass(ID), TM(TM) {
	codeAddress = CodeStartAddress;
	rodataAddress = 0; // Will be set after the doInitialization()
}

bool StaticAddressProvider::runOnMachineFunction(MachineFunction &F)
{
	auto arch = TM.getTargetTriple().getArch();
	// RISC-V has constant pools for constants of size "double"
	// They are put in a part of the rodata segment
	if (arch == llvm::Triple::ArchType::riscv32) {
		auto MCP = F.getConstantPool()->getConstants();
		for (unsigned cpeidx = 0; cpeidx < MCP.size(); ++cpeidx) {
			cpe2addr.insert(std::make_pair(std::make_pair(&F, cpeidx), rodataAddress));
			unsigned entrysize = MCP[cpeidx].getType()->getPrimitiveSizeInBits();
			assert (entrysize % 8 == 0);
			rodataAddress += (entrysize / 8);
		}
	}

	// Run on basic blocks individually
  bool Changed = false;
  for (MachineFunction::iterator FI = F.begin(); FI != F.end(); ++FI)
    Changed |= runOnMachineBasicBlock(*FI);
  return Changed;
}

bool StaticAddressProvider::runOnMachineBasicBlock(MachineBasicBlock &MBB)
{
	unsigned skipAddresses = 0; // Skip that many words in address space, e.g. for the jump table allocation

	int posInMBB = 0;

	auto arch = TM.getTargetTriple().getArch();

 	for (auto& I : MBB) {
		assert (!I.isBundle() && "We found a bundled instruction!");

		bool assignAddress = false;

		// Remember the position of the instruction within the basic block
		ins2posinbb.insert(std::make_pair(&I, posInMBB));
		++posInMBB;

		unsigned instrAllocSizeInBytes = 4;
		if (isPseudoInstruction(&I)) {
			if (arch == llvm::Triple::ArchType::arm) {
				switch (I.getOpcode()) {
					// If we have a constant pool entry remember its address
					case ARM::CONSTPOOL_ENTRY:
						{
							unsigned LabelID = I.getOperand(0).getImm(); // The label to load from
							unsigned CPI = I.getOperand(1).getIndex(); // Index of the given constant in the pool
							auto MCP = I.getParent()->getParent()->getConstantPool()->getConstants();
							assert(CPI < MCP.size() && "Invalid constant pool index.");
							auto CPELab = std::make_pair(I.getParent()->getParent(), LabelID);
							assert(cpe2addr.count(CPELab) == 0 && "We saw same constant pool entry twice");
							auto CPE = MCP[CPI].Val.ConstVal;
							// If constant pool contains an address to global variable track it
							if (const GlobalVariable* GV = dyn_cast<GlobalVariable>(CPE)) {
								unsigned varAddr = glvar2addr.at(GV);
								addr2values.insert(std::make_pair(codeAddress, varAddr));
								addr2symbol.insert(std::make_pair(codeAddress, GV));
							}
							cpe2addr.insert(std::make_pair(CPELab, codeAddress));
							assignAddress = true;
							// CP entry can allocate more than 4 bytes
							instrAllocSizeInBytes = I.getOperand(2).getImm();
							assert ((instrAllocSizeInBytes % 4 == 0) && "Cannot have non-4-divisble cp entry size");
							break;
						}
					// We ignore the pseudo-instructions that do not have any effect later on
					case ARM::IMPLICIT_DEF:
					case ARM::KILL:
					case ARM::DBG_VALUE:
					case ARM::CFI_INSTRUCTION:
						assert(I.isTransient() && "Non-effect pseudo might have effect?");
						break;
					// TODO review, why are these instructions pseudo? Have one ISA corresponding instruction...
					// We treat the following instructions as non-pseudo-instructions
					case ARM::B:
					case ARM::MULv5:
					case ARM::SMULLv5:
					case ARM::UMULLv5:
					case ARM::UMLALv5:
					case ARM::MLAv5:
					case ARM::LEApcrelJT:
					case ARM::LEApcrel:
					case ARM::BR_JTr:
					case ARM::BR_JTm_rs:
					case ARM::BR_JTm_i12:
						assignAddress = true;
						break;
					case ARM::JUMPTABLE_ADDRS:
						{
							assert (I.getOperand(1).isJTI() && "Cannot JUMPTABLE to non jump table");
							assert (MBB.getParent()->getJumpTableInfo() && "Have JUMPTABLE but not jump table");
							auto JTindex = I.getOperand(1).getIndex();
							//assert (I.getOperand(0).getImm() == JTindex && "First operand expected to be jt-index");
							auto& MJTEs = MBB.getParent()->getJumpTableInfo()->getJumpTables();
							auto& targetMBBs = MJTEs[JTindex].MBBs;
							assert (targetMBBs.size() > 0 && "Empty or negative jumptable");
							unsigned jtsize = targetMBBs.size();
							assert (I.getOperand(2).getImm() == 4*jtsize && "Third operand expected to be jt-size");
  						skipAddresses += jtsize;
							assignAddress = false;
							break;
						}
					default:
						{
							errs() << (isPseudoInstruction(&I) ? "true" : "false") << I;
							assert (0 && "We have unhandled pseudo instructions");
							break;
						}
				}
			} else if (arch == llvm::Triple::riscv32) {
				switch (I.getOpcode()) {
					// We ignore the pseudo-instructions that do not have any effect later on
					case RISCV::IMPLICIT_DEF:
					case RISCV::KILL:
					case RISCV::CFI_INSTRUCTION:
						assert(I.isTransient() && "Non-effect pseudo might have effect?");
						break;
					// TODO review, why are all instructions classified as pseudo?
					// We treat the following instructions as non-pseudo-instructions
					case RISCV::LUI:

					case RISCV::JAL:
					case RISCV::JALR:
					case RISCV::PseudoRET:
					case RISCV::PseudoCALL:
					case RISCV::PseudoBR:

					case RISCV::BEQ:
					case RISCV::BNE:
					case RISCV::BLT:
					case RISCV::BGE:
					case RISCV::BLTU:
					case RISCV::BGEU:

					case RISCV::LB:
					case RISCV::LH:
					case RISCV::LW:
					case RISCV::LBU:
					case RISCV::LHU:

					case RISCV::SB:
					case RISCV::SH:
					case RISCV::SW:

					case RISCV::ADDI:
					case RISCV::SLTI:
					case RISCV::SLTIU:
					case RISCV::XORI:
					case RISCV::ORI:
					case RISCV::ANDI:
					case RISCV::SLLI:
					case RISCV::SRLI:
					case RISCV::SRAI:

					case RISCV::ADD:
					case RISCV::SUB:
					case RISCV::SLL:
					case RISCV::SLT:
					case RISCV::SLTU:
					case RISCV::XOR:
					case RISCV::SRL:
					case RISCV::SRA:
					case RISCV::OR:
					case RISCV::AND:

					case RISCV::MUL:
					case RISCV::MULH:
					case RISCV::MULHSU:
					case RISCV::MULHU:
					case RISCV::DIV:
					case RISCV::DIVU:
					case RISCV::REM:
					case RISCV::REMU:

					case RISCV::FLW:
					case RISCV::FSW:
					case RISCV::FLD:
					case RISCV::FSD:

					case RISCV::FMADD_S:
					case RISCV::FMADD_D:
					case RISCV::FMSUB_S:
					case RISCV::FMSUB_D:
					case RISCV::FNMSUB_S:
					case RISCV::FNMSUB_D:
					case RISCV::FNMADD_S:
					case RISCV::FNMADD_D:
					case RISCV::FADD_S:
					case RISCV::FADD_D:
					case RISCV::FSUB_S:
					case RISCV::FSUB_D:
					case RISCV::FMUL_S:
					case RISCV::FMUL_D:
					case RISCV::FDIV_S:
					case RISCV::FDIV_D:

					case RISCV::FSQRT_S:
					case RISCV::FSQRT_D:

					case RISCV::FSGNJ_S:
					case RISCV::FSGNJ_D:
					case RISCV::FSGNJN_S:
					case RISCV::FSGNJN_D:
					case RISCV::FSGNJX_S:
					case RISCV::FSGNJX_D:

					case RISCV::FLE_S:
					case RISCV::FLE_D:
					case RISCV::FLT_S:
					case RISCV::FLT_D:
					case RISCV::FEQ_S:
					case RISCV::FEQ_D:

					case RISCV::FMIN_S:
					case RISCV::FMIN_D:
					case RISCV::FMAX_S:
					case RISCV::FMAX_D:

					case RISCV::FMV_W_X:
					case RISCV::FMV_X_W:

					case RISCV::FCVT_S_W:
					case RISCV::FCVT_S_WU:
					case RISCV::FCVT_W_S:
					case RISCV::FCVT_WU_S:
					case RISCV::FCVT_S_D:
					case RISCV::FCVT_D_S:
					case RISCV::FCVT_W_D:
					case RISCV::FCVT_WU_D:
					case RISCV::FCVT_D_W:
					case RISCV::FCVT_D_WU:

						assignAddress = true;
						break;
					default:
						{
							errs() << (isPseudoInstruction(&I) ? "true" : "false") << I;
							assert (0 && "We have unhandled pseudo instructions");
							break;
						}
				}
			}
		} else {
			assignAddress = true;
		}

		if (assignAddress) {
			assert(!I.isTransient() && "Instructions might have no effect, but allocate anyway?");
			assert(ins2addr.count(&I) == 0 && "We saw the same instruction twice.");
			ins2addr.insert(std::make_pair(&I, codeAddress));
			assert(addr2ins.count(codeAddress) == 0 && "We saw the same address for different instr.");
			addr2ins.insert(std::make_pair(codeAddress, &I));

			codeAddress += instrAllocSizeInBytes;
		}
 	}

	codeAddress += skipAddresses * 4;

  return false;
}

bool StaticAddressProvider::doInitialization(Module& M) {
	// TODO Take correct addresses here
	unsigned startdatasection = 0x00d0000;

	dataLayout = std::make_unique<const DataLayout>(&M);
	for (Module::const_global_iterator gvi = M.global_begin(); gvi != M.global_end(); ++gvi)
	{
		/* align the startaddress to the next variable's desired alignment */
		unsigned alignment = gvi->getAlignment();
		assert((alignment & (alignment-1)) == 0 && "alignment is not a power of 2");
		if (startdatasection & (alignment-1)) {
			startdatasection &= ~(alignment-1);
			startdatasection += alignment;
		}

		glvar2addr.insert(std::make_pair(&*gvi, startdatasection));
		DEBUG_WITH_TYPE("staddrprov", dbgs() << startdatasection << ": " << *gvi << "\n");
		auto ptrType = gvi->getType();
		assert (ptrType->isPointerTy() && "Global Variable that is not a pointer");
		auto elementType = ptrType->getPointerElementType();
		startdatasection += M.getDataLayout().getTypeAllocSize(elementType);
	}

	for (auto& gl2ad : glvar2addr)
	{
		if (!gl2ad.first->hasInitializer())
			continue;
		auto gvInit = gl2ad.first->getInitializer();
		// If global variable refers another variable, or a constant casted to ptr, track it
		if (const GlobalVariable *GV = dyn_cast<GlobalVariable>(gvInit)) {
			addr2values.insert(std::make_pair(gl2ad.second, glvar2addr.at(GV)));
			addr2symbol.insert(std::make_pair(gl2ad.second, GV));
		} else if (const ConstantExpr *CE = dyn_cast<ConstantExpr>(gvInit)) {
			// We require an inttoptr opcode with only one operand
			if (CE->getOpcode() == 43 && CE->op_begin()+1 == CE->op_end()) {
				auto& firstOp = *(CE->op_begin());
				if (const ConstantInt* CI = dyn_cast<ConstantInt>(firstOp)) {
					auto value = CI->getSExtValue();
					addr2values.insert(std::make_pair(gl2ad.second, value));
				}
			}
		}
	}
  rodataAddress = startdatasection;
	return false;
}

bool StaticAddressProvider::hasMachineInstrByAddr(unsigned addr) {
	return addr2ins.count(addr) > 0;
}

const MachineInstr* StaticAddressProvider::getMachineInstrByAddr(unsigned addr) {
	assert (addr2ins.count(addr) > 0 && "Cannot find instruction for given address");
	return addr2ins[addr];
}

unsigned StaticAddressProvider::getAddr(const MachineInstr* I) {
	assert(ins2addr.count(I) > 0 && "Cannot find address for this instruction");
	return ins2addr.find(I)->second;
}

unsigned StaticAddressProvider::getConstPoolEntryAddr(const MachineFunction* MF, unsigned idx) {
	assert(cpe2addr.count(std::make_pair(MF, idx)) > 0 && "Cannot find address for CP entry");
	return cpe2addr.find(std::make_pair(MF, idx))->second;
}

bool StaticAddressProvider::hasValueAtAddress(unsigned addr) {
	return addr2values.count(addr) > 0;
}

unsigned StaticAddressProvider::getValueAtAddress(unsigned addr) {
	assert(hasValueAtAddress(addr) && "Cannot get value at the given address");
	return addr2values.at(addr);
}

const GlobalVariable* StaticAddressProvider::getSymbolAtAddress(unsigned addr) {
	if (addr2symbol.count(addr) > 0) {
		return addr2symbol.at(addr);
	} else {
		return nullptr;
	}
}

bool StaticAddressProvider::hasGlobalVarAddress(const GlobalVariable* glvar) {
	return glvar2addr.count(glvar) > 0;
}

unsigned StaticAddressProvider::getGlobalVarAddress(const GlobalVariable* glvar) {
	assert (hasGlobalVarAddress(glvar) && "Cannot get address for given global variable");
	return glvar2addr.at(glvar);
}

unsigned StaticAddressProvider::getArraySize(const GlobalVariable *glvar)
{
	Type *type = glvar->getType()->getElementType();
	assert(type->isSized());
	unsigned size = dataLayout->getTypeAllocSize(type);
	/* if our size is not divisible by four we have to acccess the
	 * surrounding word anyway - round up to the next multiple of 4 */
	if (size % 4 != 0) {
		size = (size & ~3) + 4;
	}
	return size;
}

std::set<const GlobalVariable*> StaticAddressProvider::getGlobalVariables() const
{
	std::set<const GlobalVariable*> allkeys;
	for (auto& gva : glvar2addr) {
		allkeys.insert(gva.first);
	}
	return allkeys;
}

boost::optional<int> StaticAddressProvider::getGlobalVariableInitialValue(const GlobalVariable* glvar) const
{
	if (glvar->hasDefinitiveInitializer() && !glvar->isExternallyInitialized()) {
		auto constant = glvar->getInitializer();
		//errs() << *constant << "\n";
		if (const ConstantExpr* ce = dyn_cast<ConstantExpr>(constant)) {
			if (ce->isCast()) {
				assert ((ce->op_begin()+1 == ce->op_end()) && "Internal assumption violated");
				if (const ConstantInt* ci = dyn_cast<ConstantInt>(*ce->op_begin())) {
					if (ci->getValue().getActiveBits() <= 32) {
						int finval = (int) ci->getLimitedValue();
						return finval;
					} else {
						errs() << "[Warning] Detected a constant initializer whose value does not fit into 32 bits.\n";
					}
				}
			}
		}
	}
	return boost::none;
}

bool StaticAddressProvider::goesExternal(unsigned addr)
{
	if (this->hasMachineInstrByAddr(addr)) {
		// We have an instruction for this address, so likely it is not going extern
		return false;
	}
	// Likely we are external otherwise
	return true;
}

void StaticAddressProvider::dump(std::ostream& mystream) const {
	mystream 	<< "Address Mapping\n"
						<< "---------------\n\n";
	for (auto& kv : addr2ins) {
		printHex(mystream, kv.first);
		mystream << ": " << getMachineInstrIdentifier(kv.second) << "\n";
	}
	mystream << "\nAddress to Address Mapping\n"
					 << "------------------------\n\n";
	for (auto& av : addr2values) {
		printHex(mystream, av.first);
		mystream << ": ";
		printHex(mystream, av.second);
		if (addr2symbol.count(av.first) > 0) {
			mystream << ", @" << addr2symbol.at(av.first)->getName().str();
		}
		mystream << "\n";
	}

	mystream << "\nGlobal Variables to Address Mapping\n"
					 << "-----------------------------------\n\n";

	llvm::raw_os_ostream myllvmstream(mystream);
	if(!ArrayAnalysis) {
		myllvmstream << "Note: Array Awareness disabled\n";
	}

	for (auto& ga : glvar2addr) {
		myllvmstream << *ga.first << ": ";
		printHex(myllvmstream, ga.second);
		myllvmstream << "\n";
	}
}

}

FunctionPass *llvm::createStaticAddressProvider(TargetMachine &TM) {
  TimingAnalysisPass::staticAddressProvider = new TimingAnalysisPass::StaticAddressProvider(TM);
  return TimingAnalysisPass::staticAddressProvider;
}

