////////////////////////////////////////////////////////////////////////////////
//
//   LLVMTA - Timing Analyser performing worst-case execution time analysis
//     using the LLVM backend intermediate representation
//
// Copyright (C) 2013-2015  Saarland University
// Copyright (C) 2014-2015  Claus Faymonville
//
// This file is distributed under the Saarland University Software Release
// License. See LICENSE.TXT for details.
//
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING BUT NOT LIMITED TO, THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE SAARLAND UNIVERSITY, THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING BUT NOT LIMITED TO PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED OR OTHER LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, TORT OR OTHERWISE, ARISING IN ANY WAY FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH A DAMAGE.
//
////////////////////////////////////////////////////////////////////////////////

#include "LLVMPasses/TimingAnalysisMain.h"
#include "LLVMPasses/MachineFunctionCollector.h"
#include "PartitionUtil/DirectiveHeuristics.h"
#include "AnalysisFramework/AnalysisDriver.h"
#include "AnalysisFramework/CallGraph.h"
#include "AnalysisFramework/AnalysisResults.h"

#include "PreprocessingAnalysis/ConstantValueDomain.h"
#include "PreprocessingAnalysis/AddressInformation.h"
#include "Memory/PersistenceScopeInfo.h"

#include "PathAnalysis/LoopBoundInfo.h"

#include "LLVMPasses/DispatchInOrderPipeline.h"
#include "LLVMPasses/DispatchPretPipeline.h"
#include "LLVMPasses/DispatchOutOfOrderPipeline.h"
#include "LLVMPasses/DispatchFixedLatency.h"
#include "LLVMPasses/DispatchMemory.h"

#include "Util/Options.h"
#include "Util/Statistics.h"

#include "llvm/CodeGen/MachineFunctionPass.h"
#include "llvm/CodeGen/Passes.h"
#include "llvm/Support/Format.h"

#include <fstream>
#include <type_traits>
#include <sstream>
#include <list>
#include <utility>
#include <cmath>
#include <limits>

using namespace llvm;
using namespace std;

namespace TimingAnalysisPass {

unsigned getInitialStackPointer() {
	return InitialStackPointer;
}

unsigned getInitialLinkRegister() {
	return InitialLinkRegister;
}

MachineFunction* getAnalysisEntryPoint() {
	auto res = machineFunctionCollector->getFunctionByName(AnalysisEntryPoint);
	assert (res && "Invalid entry point specified");
	return res;
}

char TimingAnalysisMain::ID = 0;
TargetMachine* TimingAnalysisMain::targetMachine = nullptr;
    
TimingAnalysisMain::TimingAnalysisMain(TargetMachine &TM) 
	: MachineFunctionPass(ID)
{ 
	targetMachine = &TM;
}

TargetMachine& TimingAnalysisMain::getTargetMachine() {
	return *targetMachine;
}
		
bool TimingAnalysisMain::runOnMachineFunction(MachineFunction &MF) {
  bool Changed = false;
  return Changed;
}

bool TimingAnalysisMain::doFinalization(Module& M)
{
	if (!machineFunctionCollector->hasFunctionByName(AnalysisEntryPoint)) {
		outs() << "No Timing Analysis Run. There is no entry point: " << AnalysisEntryPoint << "\n";
		exit(1);
	}

	ofstream myfile;

	// Default analysis type: timing
	if (AnaType.getBits() == 0) {
		AnaType.addValue(AnalysisType::TIMING);
	}

	Statistics& stats = Statistics::getInstance();
	stats.startMeasurement("Complete Analysis");

	if (CoRunnerSensitive) {
		for (int i = 0; i <= UntilIterationMeasurement; ++i) {
			std::string measurementId = "Until Iteration ";
			measurementId += std::to_string(i);
			stats.startMeasurement(measurementId);
		}
	}

	if (OutputExtFuncAnnotationFile) {
		myfile.open("ExtFuncAnnotations.csv", ios_base::trunc);
		CallGraph::getGraph().dumpUnknownExternalFunctions(myfile);
		myfile.close();
		return false;
	}

	outs() << "Timing Analysis for entry point: " << AnalysisEntryPoint << "\n";

	if (!QuietMode)
	{
		myfile.open ("AnnotatedHeuristics.txt", ios_base::trunc);
		directiveHeuristicsPass->dump(myfile);
		myfile.close();

		myfile.open ("PersistenceScopes.txt", ios_base::trunc);
		PersistenceScopeInfo::getInfo().dump(myfile);
		myfile.close();

		myfile.open ("CallGraph.txt", ios_base::trunc);
		CallGraph::getGraph().dump(myfile);
		myfile.close();
	}

	VERBOSE_PRINT(" -> Finished Preprocessing Phase\n");

	// Dispatch the value analysis
	auto arch = getTargetMachine().getTargetTriple().getArch();
	if (arch == Triple::ArchType::arm) {
		dispatchValueAnalysis<Triple::ArchType::arm>();
	} else if (arch == Triple::ArchType::riscv32) {
		dispatchValueAnalysis<Triple::ArchType::riscv32>();
	} else {
		assert (0 && "Unsupported ISA for LLVMTA");
	}

	return false;
}

template <Triple::ArchType ISA>
void TimingAnalysisMain::dispatchValueAnalysis()
{
	ofstream myfile;

	std::tuple<> noDep;
	AnalysisDriverInstr<ConstantValueDomain<ISA>> constValAna(AnalysisEntryPoint, noDep);
	auto cvAnaInfo = constValAna.runAnalysis();

	loopBoundInfoPass->computeLoopBoundFromCVDomain(*cvAnaInfo);

	if (OutputLoopAnnotationFile) {
		ofstream myfile2;
		myfile.open("CtxSensLoopAnnotations.csv", ios_base::trunc);
		myfile2.open("LoopAnnotations.csv", ios_base::trunc);
		loopBoundInfoPass->dumpNonUpperBoundLoops(myfile, myfile2);
		myfile2.close();
		myfile.close();
		return;
	}

	for (auto BoundsFile : ManualLoopBounds) {
		loopBoundInfoPass->parseManualUpperLoopBounds(BoundsFile.c_str());
	}

	if (!QuietMode)
	{
		myfile.open ("LoopBounds.txt", ios_base::trunc);
		loopBoundInfoPass->dump(myfile);
		myfile.close();

		myfile.open ("ConstantValueAnalysis.txt", ios_base::trunc);
		cvAnaInfo->dump(myfile);
		myfile.close();
	}

  AddressInformationImpl<ConstantValueDomain<ISA>> addrInfo(*cvAnaInfo);

  if (!QuietMode)
  {
		myfile.open ("AddressInformation.txt", ios_base::trunc);
		addrInfo.dump(myfile);
		myfile.close();
	}

	VERBOSE_PRINT(" -> Finished Value & Address Analysis\n");

	// Set and check the parameters of the instruction and data cache
	icacheConf.LINE_SIZE     = Ilinesize;
	icacheConf.ASSOCIATIVITY = Iassoc;
	icacheConf.N_SETS        = Insets;
	icacheConf.checkParams();

	dcacheConf.LINE_SIZE     = Dlinesize;
	dcacheConf.ASSOCIATIVITY = Dassoc;
	dcacheConf.N_SETS        = Dnsets;
	dcacheConf.WRITEBACK     = DataCacheWriteBack;
	dcacheConf.WRITEALLOCATE = DataCacheWriteAllocate;
	dcacheConf.checkParams();

	// Select the analysis to execute
	dispatchAnalysisType(addrInfo);

	// No need for constant value information
	delete cvAnaInfo;
	// Release the call graph instance
	CallGraph::getGraph().releaseInstance();

	// Write results and statistics
	Statistics& stats = Statistics::getInstance();
	AnalysisResults& ar = AnalysisResults::getInstance();

	stats.stopMeasurement("Complete Analysis");

	myfile.open("Statistics.txt", ios_base::trunc);
	stats.dump(myfile);
	myfile.close();

	myfile.open("TotalBound.xml", ios_base::trunc);
	ar.dump(myfile);
	myfile.close();
}

void TimingAnalysisMain::dispatchAnalysisType(AddressInformation& addressInfo)
{
	AnalysisResults& ar = AnalysisResults::getInstance();
	// Timing & CRPD calculation need normal muarch analysis first
	if (AnaType.isSet(AnalysisType::TIMING) || AnaType.isSet(AnalysisType::CRPD)) {
		auto bound = dispatchTimingAnalysis(addressInfo);
		ar.registerResult("total", bound);
		if (bound) {
			outs() << "Calculated Timing Bound: " << llvm::format("%-20.0f", bound.get().ub) << "\n";
		} else {
			outs() << "Calculated Timing Bound: infinite\n";
		}
	}
	if (AnaType.isSet(AnalysisType::L1ICACHE))
	{
		auto bound = dispatchCacheAnalysis(AnalysisType::L1ICACHE, addressInfo);
		ar.registerResult("icache", bound);
		if (bound) {
			outs() << "Calculated "
					<< "Instruction Cache Miss Bound: " << llvm::format("%-20.0f", bound.get().ub) << "\n";
		} else {
			outs() << "Calculated "
					<< "Instruction Cache Miss Bound: infinite\n";
		}
	}
	if (AnaType.isSet(AnalysisType::L1DCACHE))
	{
		auto bound = dispatchCacheAnalysis(AnalysisType::L1DCACHE, addressInfo);
		ar.registerResult("dcache", bound);
		if (bound) {
			outs() << "Calculated "
					<< "Data Cache Miss Bound: " << llvm::format("%-20.0f", bound.get().ub) << "\n";
		} else {
			outs() << "Calculated "
					<< "Data Cache Miss Bound: infinite\n";
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
/// Timing Analysis ///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

boost::optional<BoundItv> TimingAnalysisMain::dispatchTimingAnalysis(AddressInformation& addressInfo)
{
	switch (MuArchType) {
		case MicroArchitecturalType::FIXEDLATENCY:
			assert (MemTopType == MemoryTopologyType::NONE && "Fixed latency has no external memory");
			return dispatchFixedLatencyTimingAnalysis();
		case MicroArchitecturalType::PRET:
			return dispatchPretTimingAnalysis(addressInfo);
		case MicroArchitecturalType::INORDER:
		case MicroArchitecturalType::STRICTINORDER:
			return dispatchInOrderTimingAnalysis(addressInfo);
		case MicroArchitecturalType::OUTOFORDER:
			return dispatchOutOfOrderTimingAnalysis(addressInfo);
		default:
			errs() << "No known microarchitecture chosen.\n";
			return boost::none;
	}
}

boost::optional<BoundItv> TimingAnalysisMain::dispatchCacheAnalysis(AnalysisType anatype, AddressInformation& addressInfo)
{
	switch (MuArchType) {
		case MicroArchitecturalType::INORDER:
		case MicroArchitecturalType::STRICTINORDER:
			return dispatchInOrderCacheAnalysis(anatype, addressInfo);
		default:
			errs() << "Unsupported microarchitecture for standalone cache analysis.\n";
			return boost::none;
	}
}

}

FunctionPass *llvm::createTimingAnalysisMain(TargetMachine &TM) {
  return new TimingAnalysisPass::TimingAnalysisMain(TM);
}
