////////////////////////////////////////////////////////////////////////////////
//
//   LLVMTA - Timing Analyser performing worst-case execution time analysis
//     using the LLVM backend intermediate representation
//
// Copyright (C) 2013-2015  Saarland University
//
// This file is distributed under the Saarland University Software Release
// License. See LICENSE.TXT for details.
//
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING BUT NOT LIMITED TO, THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE SAARLAND UNIVERSITY, THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING BUT NOT LIMITED TO PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED OR OTHER LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, TORT OR OTHERWISE, ARISING IN ANY WAY FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH A DAMAGE.
//
////////////////////////////////////////////////////////////////////////////////

#include "PartitionUtil/DirectiveHeuristics.h"
#include "PartitionUtil/PartitionToken.h"

#include "LLVMPasses/TimingAnalysisMain.h"

#include "Util/Options.h"

#include <set>
#include <vector>

using namespace llvm;
using namespace std;

namespace TimingAnalysisPass {

DirectiveHeuristicsPass* directiveHeuristicsPass;

char DirectiveHeuristicsPass::ID = 0;

DirectiveHeuristicsPass::DirectiveHeuristicsPass() 
			: MachineFunctionPass(ID) { }

void DirectiveHeuristicsPass::getAnalysisUsage(AnalysisUsage &AU) const {
  MachineFunctionPass::getAnalysisUsage(AU);
  AU.setPreservesCFG();
  AU.addRequired<MachineLoopInfo>();
}

////////////////////////////////////////////////////////////////////////////
/// Directive Annotation on Function, Basic Block, and Instruction Level ///
////////////////////////////////////////////////////////////////////////////

bool DirectiveHeuristicsPass::runOnMachineFunction(MachineFunction &MF) {
	// The partitioning on the callee site (added on each call)
	shared_ptr<PartitionToken> ptCallee(new PartitionTokenFunCallee(&MF));
	set<shared_ptr<PartitionToken>> ptsetCallee;
	ptsetCallee.insert(ptCallee);
	Directive* direcCreateCallee = new Directive(DirectiveType::CREATE, ptsetCallee);
	Directive* direcMergeCallee = new Directive(DirectiveType::MERGE, ptsetCallee);
	this->directiveBeforeFunc.insert(make_pair(&MF , direcCreateCallee));
	this->directiveAfterFunc.insert(make_pair(&MF, direcMergeCallee));

	// If Loop Peeling is desired, we do it
	if (LoopPeel > 0) {
		// Iterate over all level 1 loops and let them annotate
		MachineLoopInfo &MLI = getAnalysis<MachineLoopInfo>();
		for (auto machineLoop : MLI) {
			annotateLoopDirective(machineLoop);
		}
	}

	// Derive Directives on Instruction Level (e.g. Function Call)
	for (MachineFunction::iterator MBB = MF.begin(); MBB != MF.end(); ++MBB) {
		this->runOnMachineBasicBlock(*MBB);
	}
	return false; // Nothing is changed, no fixed-point iteration necessary
}

void DirectiveHeuristicsPass::annotateLoopDirective(const MachineLoop* loop) {
	// Create PartitionTokens
	auto backedgeSet = getBackEdgesOfLoop(loop);
	set<shared_ptr<PartitionToken>> ptset;
	for (unsigned i = 0; i < LoopPeel; ++i) {
		shared_ptr<PartitionToken> ptLoopPeel(new PartitionTokenLoopPeel(loop, backedgeSet, i));
		ptset.insert(ptLoopPeel);
	}
	shared_ptr<PartitionToken> ptLoopIter(new PartitionTokenLoopIter(loop, backedgeSet, LoopPeel));
	ptset.insert(ptLoopIter);
	// Add create directive at last possible position on incoming edges
	auto incomingEdges = getIncomingEdgesOfLoop(loop);
	for (auto edge : incomingEdges) {
		if (directiveLeaveBBEdge.count(edge) == 0) {
			list<Directive*>* newlist = new list<Directive*>();
			directiveLeaveBBEdge.insert(make_pair(edge, newlist));
		}
		auto edgeDirectiveList = directiveLeaveBBEdge[edge];
		Directive* direcLoopCreate = new Directive(DirectiveType::CREATE, ptset);
		edgeDirectiveList->push_back(direcLoopCreate);
	}
	// Add merge directive at first possible position on outgoing edges
	auto exitingEdges = getExitingEdgesOfLoop(loop);
	for (auto edge : exitingEdges) {
		if (directiveEnterBBEdge.count(edge) == 0) {
			list<Directive*>* newlist = new list<Directive*>();
			directiveEnterBBEdge.insert(make_pair(edge, newlist));
		}
		auto edgeDirectiveList = directiveEnterBBEdge[edge];
		Directive* direcLoopMerge = new Directive(DirectiveType::MERGE, ptset);
		edgeDirectiveList->push_front(direcLoopMerge);
	}
	// Do directive annotations for all nested loops
	for (auto subloop : loop->getSubLoops()) {
		annotateLoopDirective(subloop);
	}
}

bool DirectiveHeuristicsPass::runOnMachineBasicBlock(MachineBasicBlock &MBB) {
	// We add partitioning directives for function calls around a call instruction.
	for (auto& MI : MBB) {
		// Check if current instruction is a calling instructions
		if(MI.isCall()) { // TODO: how are computed calls handled?
			// Get the call target operand
			MachineOperand& callTarget = MI.getOperand(0);
			// We do not support offset function calls
			assert(callTarget.getOffset() == 0 && "No offsets allowed for function calls");
	
			// Create corresponding directives
			// The partitioning on the caller site here (partitioning on the callee site above)
			shared_ptr<PartitionToken> pt(new PartitionTokenCallSite(&MI));
			set<shared_ptr<PartitionToken>> ptset;
			ptset.insert(pt);
			Directive* direcCreate = new Directive(DirectiveType::CREATE, ptset);
			Directive* direcMerge = new Directive(DirectiveType::MERGE, ptset);
			this->directiveBeforeInstr.insert(make_pair(&MI, direcCreate));
			this->directiveAfterInstr.insert(make_pair(&MI, direcMerge));
		}
	}
	return false;
}

///////////////////
/// Output dump ///
///////////////////

void DirectiveHeuristicsPass::dump(std::ostream& mystream) const {
	mystream << "##################################\n"
						<< "## Directives Machine Functions ##\n"
						<< "##################################\n";
	// Before function invocation
	for (auto it = directiveBeforeFunc.begin(); it != directiveBeforeFunc.end(); ++it) {
		mystream << "Before " << it->first->getName().str() << ", " << *(it->second) << "\n";
	}
	// After function invocation
	for (auto it = directiveAfterFunc.begin(); it != directiveAfterFunc.end(); ++it) {
		mystream << "After " << it->first->getName().str() << ", " << *(it->second) << "\n";
	}

 	mystream << "#############################\n"
						<< "## Directives Basic Blocks ##\n"
						<< "#############################\n";
	// Entering basic block edge
	for (auto it = directiveEnterBBEdge.begin(); it != directiveEnterBBEdge.end(); ++it) {
		mystream << "Enter (" << getBasicBlockIdentifier(it->first.first) << ", " 
											<< getBasicBlockIdentifier(it->first.second) << "), [";
		for (auto listit = it->second->begin(); listit != it->second->end(); ++listit) {
			mystream << *(*listit);
			if(distance(listit, it->second->end()) > 1) {
				mystream << ",\n";
			}
		}
		mystream << "]\n";
	}
	// Leaving basic block edge
	for (auto it = directiveLeaveBBEdge.begin(); it != directiveLeaveBBEdge.end(); ++it) {
		mystream << "Leave (" << getBasicBlockIdentifier(it->first.first) << ", " 
											<< getBasicBlockIdentifier(it->first.second) << "), ["; 
		for (auto listit = it->second->begin(); listit != it->second->end(); ++listit) {
			mystream << *(*listit);
			if(distance(listit, it->second->end()) > 1) {
				mystream << ",\n";
			}
		}
		mystream << "]\n";
	}

	mystream << "#############################\n"
						<< "## Directives Instructions ##\n"
						<< "#############################\n";
	// Directives before machine instructions
	for (auto it = directiveBeforeInstr.begin(); it != directiveBeforeInstr.end(); ++it) {
		mystream << "Before " << getMachineInstrIdentifier(it->first) << ", " << *(it->second) << "\n";
	}
	// Directives after machine instructions
	for (auto it = directiveAfterInstr.begin(); it != directiveAfterInstr.end(); ++it) {
		mystream << "After " << getMachineInstrIdentifier(it->first) << ", " << *(it->second) << "\n";
	}
}

///////////////////////////////////////
/// Access the resulting directives ///
///////////////////////////////////////

bool DirectiveHeuristicsPass::hasDirectiveBeforeInstr(const MachineInstr* MI) {
	return this->directiveBeforeInstr.count(MI) > 0;
}

bool DirectiveHeuristicsPass::hasDirectiveAfterInstr(const MachineInstr* MI) {
	return this->directiveAfterInstr.count(MI) > 0;
}

bool DirectiveHeuristicsPass::hasDirectiveOnCall(const MachineFunction* MF) {
	return this->directiveBeforeFunc.count(MF) > 0;	
}

bool DirectiveHeuristicsPass::hasDirectiveOnReturn(const MachineFunction* MF) {
	return this->directiveAfterFunc.count(MF) > 0;
}

bool DirectiveHeuristicsPass::hasDirectiveOnEdgeEnter(pair<const MachineBasicBlock*, const MachineBasicBlock*> E) {
	return this->directiveEnterBBEdge.count(E) > 0;
}

bool DirectiveHeuristicsPass::hasDirectiveOnEdgeLeave(pair<const MachineBasicBlock*, const MachineBasicBlock*> E) {
	return this->directiveLeaveBBEdge.count(E) > 0;
}

Directive* DirectiveHeuristicsPass::getDirectiveBeforeInstr(const MachineInstr* MI) {
	assert(hasDirectiveBeforeInstr(MI) && "No directive found, use proper API.");
	return this->directiveBeforeInstr[MI];
}

Directive* DirectiveHeuristicsPass::getDirectiveAfterInstr(const MachineInstr* MI) {
	assert(hasDirectiveAfterInstr(MI) && "No directive found, use proper API.");
	return this->directiveAfterInstr[MI];
}

Directive* DirectiveHeuristicsPass::getDirectiveOnCall(const MachineFunction* MF) {
	assert(hasDirectiveOnCall(MF) && "No directive found, use proper API.");
	return this->directiveBeforeFunc[MF];
}

Directive* DirectiveHeuristicsPass::getDirectiveOnReturn(const MachineFunction* MF) {
	assert(hasDirectiveOnReturn(MF) && "No directive found, use proper API.");
	return this->directiveAfterFunc[MF];
}

list<Directive*>* DirectiveHeuristicsPass::getDirectiveOnEdgeEnter(
						pair<const MachineBasicBlock*, const MachineBasicBlock*> E) {
	assert(hasDirectiveOnEdgeEnter(E) && "No directive found, use proper API.");
	return this->directiveEnterBBEdge[E];
}

list<Directive*>* DirectiveHeuristicsPass::getDirectiveOnEdgeLeave(
						pair<const MachineBasicBlock*, const MachineBasicBlock*> E) {
	assert(hasDirectiveOnEdgeLeave(E) && "No directive found, use proper API.");
	return this->directiveLeaveBBEdge[E];
}

}

namespace llvm {

FunctionPass* createDirectiveHeuristicsPass() {
	TimingAnalysisPass::directiveHeuristicsPass = new TimingAnalysisPass::DirectiveHeuristicsPass();
	return TimingAnalysisPass::directiveHeuristicsPass;
}

}
