////////////////////////////////////////////////////////////////////////////////
//
//   LLVMTA - Timing Analyser performing worst-case execution time analysis
//     using the LLVM backend intermediate representation
//
// Copyright (C) 2013-2015  Saarland University
// Copyright (C) 2014-2015  Claus Faymonville
//
// This file is distributed under the Saarland University Software Release
// License. See LICENSE.TXT for details.
//
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY EXPRESSED OR IMPLIED
// WARRANTIES, INCLUDING BUT NOT LIMITED TO, THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE SAARLAND UNIVERSITY, THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING BUT NOT LIMITED TO PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED OR OTHER LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, TORT OR OTHERWISE, ARISING IN ANY WAY FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH A DAMAGE.
//
////////////////////////////////////////////////////////////////////////////////

#include "PathAnalysis/LoopBoundInfo.h"

#include "LLVMPasses/MachineFunctionCollector.h"
#include "LLVMPasses/TimingAnalysisMain.h"
#include "AnalysisFramework/CallGraph.h"
#include "Util/Util.h"
#include "Util/Options.h"

#include "llvm/Analysis/ScalarEvolution.h"
#include "llvm/Analysis/ScalarEvolutionExpressions.h"
#include "llvm/CodeGen/MachineBasicBlock.h"
#include "llvm/CodeGen/MachineFunction.h"
#include "llvm/CodeGen/MachineInstr.h"

#include "llvm/Support/Debug.h"
#include "llvm/Support/raw_os_ostream.h"

#include <vector>
#include <boost/tokenizer.hpp>
#include <fstream>

using namespace llvm;

namespace TimingAnalysisPass {

LoopBoundInfoPass* loopBoundInfoPass;

char LoopBoundInfoPass::ID = 0;

LoopBoundInfoPass::LoopBoundInfoPass()
			: MachineFunctionPass(ID) {
}

void LoopBoundInfoPass::getAnalysisUsage(AnalysisUsage &AU) const {
  MachineFunctionPass::getAnalysisUsage(AU);
  AU.setPreservesCFG();
  AU.addRequired<LoopInfoWrapperPass>();
  AU.addRequired<MachineLoopInfo>();
  AU.addRequired<ScalarEvolutionWrapperPass>();
}

bool LoopBoundInfoPass::runOnMachineFunction(MachineFunction &MF) {
	LoopInfo &LI = getAnalysis<LoopInfoWrapperPass>().getLoopInfo();
	for (auto loop : LI) {
		walkLoop(loop);
	}
	MachineLoopInfo &MLI = getAnalysis<MachineLoopInfo>();
	for (auto machineLoop : MLI) {
		walkMachineLoop(machineLoop);
	}
  return false;
}

void LoopBoundInfoPass::walkLoop(const Loop* loop) {
	irLoops.push_back(loop);
	DEBUG_WITH_TYPE("loopbound", dbgs() << "***** We found a loop *****\n");
	DEBUG_WITH_TYPE("loopbound", dbgs() << "Begin subloops\n");
	for (auto subloop : loop->getSubLoops()) {
		walkLoop(subloop);
	}
	DEBUG_WITH_TYPE("loopbound", dbgs() << "End subloops\n");
}

bool LoopBoundInfoPass::isMachineLoopExactMatch(const MachineLoop* maloop, const Loop* irloop)
{
	auto irloopBBs = irloop->getBlocks();
	auto maloopBBs = maloop->getBlocks();
	bool foundAllirloopBB = true;
	for (const BasicBlock* irBB : irloopBBs) {
		bool found = false;
		for (const MachineBasicBlock* maBB : maloopBBs) {
			found |= (irBB == maBB->getBasicBlock());
		}
		foundAllirloopBB &= found;
	}
	bool foundAllmaloopBB = true;
	for (const MachineBasicBlock* maBB : maloopBBs) {
		bool found = false;
		for (const BasicBlock* irBB : irloopBBs) {
			found |= (maBB->getBasicBlock() == irBB);
		}
		foundAllmaloopBB &= found;
	}
	return foundAllirloopBB && foundAllmaloopBB;
}

bool LoopBoundInfoPass::isMachineLoopPartialMatch(const MachineLoop* maloop, const Loop* irloop)
{
	DEBUG_WITH_TYPE("loopmatch", dbgs() << "Try to match with " << *irloop);
	auto irloopBBs = irloop->getBlocks();
	auto maloopBBs = maloop->getBlocks();
	bool foundAllmaloopBB = true;
	for (const MachineBasicBlock* maBB : maloopBBs) {
		bool found = false;
		for (const BasicBlock* irBB : irloopBBs) {
			// Either there is no corresponding ir block or it matches the irloop
			found |= maBB->getBasicBlock() == nullptr;
			found |= maBB->getBasicBlock() == irBB;
		}
		DEBUG_WITH_TYPE("loopmatch", if (!found) {dbgs() << "No match for " << *maBB << " found\n";});
		foundAllmaloopBB &= found;
	}
	return foundAllmaloopBB;
}

void LoopBoundInfoPass::addSCEVMapping(const MachineLoop *ML, const Loop *loop) {
	ScalarEvolution& SEInfo = getAnalysis<ScalarEvolutionWrapperPass>().getSE();
	DEBUG_WITH_TYPE("loopbound", dbgs() << "Adding map for loop: " << *ML << "\n");
	if (SEInfo.hasLoopInvariantBackedgeTakenCount(loop)) {
		const SCEV *takenCount = SEInfo.getBackedgeTakenCount(loop);
		DEBUG_WITH_TYPE("loopbound", dbgs() << "Loop SCEV: " << *takenCount << "\n");
		DEBUG_WITH_TYPE("loopbound", dbgs() << "Loop SCEV: " << takenCount << "\n");
		upperLoopBoundsSCEV.insert(std::make_pair(ML, takenCount));
		lowerLoopBoundsSCEV.insert(std::make_pair(ML, takenCount));
	} else {
		const SCEV *maxTakenCount = SEInfo.getMaxBackedgeTakenCount(loop);
		DEBUG_WITH_TYPE ("loopbound", dbgs() << "Non-invariant Loop Bound: " << *maxTakenCount << "\n");
		// FIXME do not use max backedge taken count since it returns ridiculous numbers with llvm7 for some of our tests
		//upperLoopBoundsSCEV.insert(std::make_pair(ML, maxTakenCount));
		// We don't add the lowerLoopBound here. The SCEV here is not for
		// an exact match on the loop bound, but only offers an upper bound
		// on the number of loop iterations. We could create a new SCEV
		// with a constant value of 0 for the lower bound. That would be
		// good for symmetry. However, creating a SCEV Constant at this
		// point in the code is a very convoluted process and not worth the
		// added complexity.
	}
}

void LoopBoundInfoPass::walkMachineLoop(const MachineLoop* loop) {
	maLoops.push_back(loop);
	DEBUG_WITH_TYPE("loopmatch", dbgs() << "###### We found a machine loop #####\n" << *loop);
	std::set<const Loop*> candidates;
	for (auto irloop : irLoops) {
		// Is irloop a match?
		if (isMachineLoopPartialMatch(loop, irloop)) {
			bool nestedMatches = false;
			// Is a nested loop of irloop also a match?
			for (auto subloop : irloop->getSubLoops()) {
				nestedMatches |= isMachineLoopPartialMatch(loop, subloop);
			}
			if (!nestedMatches) {
				candidates.insert(irloop);
			}
		}
	}

	if (candidates.size() == 0) {
		DEBUG_WITH_TYPE("loopmatch", dbgs() << "[Warning] We did not find matching loop\n");
	} else if (candidates.size() == 1) {
		const Loop* correspondingLoop = *(candidates.begin());
		DEBUG_WITH_TYPE("loopmatch", dbgs() << "We found corresponding loop\n");
		if (!isMachineLoopExactMatch(loop, correspondingLoop)) {
			DEBUG_WITH_TYPE("loopmatch", dbgs() << "[Warning] We only found subset match, maybe due to optimizations.\n");
		}
		loopMapping.insert(std::make_pair(loop, correspondingLoop));
		addSCEVMapping(loop, correspondingLoop);
	} else {
		assert(0 && "[Warning] We did find too many matching loops\n");
	}

	for (auto subloop : loop->getSubLoops()) {
		walkMachineLoop(subloop);
	}
}

template <>
bool LoopBoundInfoPass::getValueByArgument<Triple::ArchType::riscv32>(
	const Argument* arg,
	const ConstantValueDomain<Triple::ArchType::riscv32>& CVDom,
	unsigned& value)
{
	// TODO riscv implementation of context sensitive loop bounds...
	return false;
}

template <>
bool LoopBoundInfoPass::getValueByArgument<Triple::ArchType::arm>(
	const Argument* arg,
	const ConstantValueDomain<Triple::ArchType::arm>& CVDom,
	unsigned& value)
{
	// Find the gpr number this argument maps to
	unsigned argnum = 0;

	// Catch information about the floating point ABI
	auto& targetOptions = TimingAnalysisMain::getTargetMachine().Options;
	auto& floatABI = targetOptions.FloatABIType;
	assert (floatABI == FloatABI::Hard || floatABI == FloatABI::Soft);

	// Scan all function arguments
	for (auto& funcarg : arg->getParent()->args()) {
		DEBUG_WITH_TYPE("loopbound", dbgs() << "Next function argument: " << funcarg << "\n");
		// Found argument. Later ones don't matter for gpr number
		if (funcarg.getArgNo() == arg->getArgNo()) {
			break;
		}

		// Get the type of argument
		auto argType = funcarg.getType();

		// Integer types occupy GPR registers
		if (argType->isIntegerTy()) {
			unsigned size = argType->getPrimitiveSizeInBits();
			assert ((size <= 32 || size == 64) && "Unexpected integer size");
			if (size <= 32) {
				++argnum;
			} else {
				// Larger integers need multiple, aligned registers
				argnum = argnum + 2 + (argnum % 2);
			}
		} else if (argType->isFloatingPointTy()) {
			// Floats have dedicated registers in hard floating point
			// mode
			if (floatABI != FloatABI::Hard) {
				if (argType->isFloatTy()) {
					++argnum;
				} else {
					assert (argType->isDoubleTy() && "Unexpected floating point type");
					argnum = argnum + 2 + (argnum % 2);
				}
			}
		} else if (argType->isPointerTy()) {
			++argnum;
		} else {
			assert (0 && "Unknown argument type");
		}
	}
	DEBUG_WITH_TYPE("loopbound", dbgs() << "Argument number: " << arg->getArgNo() << " GPR Number: " << argnum << "\n");

	// Try to get a value for the argument register
	int regnr;
	/* Accroding to the ARM ABI, registers R0-R3 hold the first 4 arguments
	 * passed to a function. Hence, we try to get those first. Other arguments
	 * are then passed through the stack. So in case of higher numbered
	 * arguments, we must translate the argument number to the location on the
	 * stack where it should be and then read the value of the constant value
	 * domain at that memory location.
	 */
	switch (argnum) {
		case 0:
			regnr = ARM::R0;
			break;
		case 1:
			regnr = ARM::R1;
			break;
		case 2:
			regnr = ARM::R2;
			break;
		case 3:
			regnr = ARM::R3;
			break;
		default:
			DEBUG_WITH_TYPE("loopbound", dbgs() << "argnum >= 4. Need to find data from Stack.\n");
			AnalysisResults::getInstance().incrementResult("SCEV_argument_high");
			regnr = -1;
	}

	if (regnr >= 0) {
		if(!(CVDom.isBottom() || CVDom.isTop(regnr))) {
			value = CVDom.getValueFor(regnr);
			return true;
		}
        DEBUG_WITH_TYPE("loopbound", dbgs() << argnum << " not found in CV Domain\n");
		DEBUG_WITH_TYPE("loopbound", dbgs() << CVDom.print());
		AnalysisResults::getInstance().incrementResult("SCEV_arg_cv");
	}

	return false;
}

template <llvm::Triple::ArchType ISA>
bool LoopBoundInfoPass::getSCEVBoundFromCVDomain(const SCEV *equation, const ConstantValueDomain<ISA>& CVDom, unsigned& value) {
	bool ret = true;
	DEBUG_WITH_TYPE ("loopbound", dbgs() << "Computing for scev: " << *equation << "\n");
	switch (static_cast<SCEVTypes>(equation->getSCEVType())) {
		case scConstant: {
			DEBUG_WITH_TYPE("loopbound", dbgs() << "Type: Constant\n");
			int64_t tmp_value = cast<SCEVConstant>(equation)->getValue()->getSExtValue();
			if (tmp_value >= std::numeric_limits<int>::max() || tmp_value <= std::numeric_limits<int>::min()) {
				DEBUG_WITH_TYPE("loopbound", dbgs() << "Skipping too large constant\n");
				AnalysisResults::getInstance().incrementResult("SCEV_constant");
				ret = false;
			} else {
				// Cut the 64-bit integer down to 32-bit (range check above) and then convert it to our container format "unsigned"
				value = (unsigned) ((int) tmp_value);
			}
			}
			break;
		case scAddExpr:
		case scMulExpr:
		case scUMaxExpr:
		case scSMaxExpr: {
			DEBUG_WITH_TYPE("loopbound", dbgs() << "Type: Nary\n");
			const SCEVNAryExpr *NAry = cast<SCEVNAryExpr>(equation);
			unsigned final_val;
			for (auto& I : NAry->operands()) {
				if (I == *NAry->op_begin()) {
					unsigned tmp;
					// Copy first value as-is
					if (!(ret = getSCEVBoundFromCVDomain(I, CVDom, tmp)))
						goto end_switch;
					final_val = tmp;
					continue;
				}

			    unsigned op_val;
				if (getSCEVBoundFromCVDomain(I, CVDom, op_val) == false) {
					ret = false;
					goto end_switch;
				}

				switch (NAry->getSCEVType()) {
					case scAddExpr:
						DEBUG_WITH_TYPE("loopbound", dbgs() << "Type: Nary (ADD)\n");
						final_val += op_val;
						break;
					case scMulExpr:
						DEBUG_WITH_TYPE("loopbound", dbgs() << "Type: Nary (Mul)\n");
#if 0
						/*
						 * Check if the multiplication can be done without
						 * overflows. If it overflows, we conservatively claim
						 * to have no knowledge about the result. This is
						 * because LLVM specifies <nsw> and <nuw> properties
						 * but doesn't state anything about what to do when an
						 * overflow happens. We just skip out in case of all
						 * overflows
						 *
						 * 04.06.2018 - Darshit Shah
						 * This code is currently commented out till we can
						 * figure out what exactly the nsw and nuw flags mean.
						 * Also, to anyone touching this in the future, make
						 * these overflow checks work with Add as well
						 */
						if (__builtin_mul_overflow_p (final_val, op_val, (unsigned) 0) == false) {
							ret = false;
							AnalysisResults::getInstance().incrementResult("SCEV/overflow");
							goto end_switch;
						}
#endif
						final_val *= op_val;
						break;
					case scUMaxExpr:
						DEBUG_WITH_TYPE("loopbound", dbgs() << "Type: Nary (UMax)\n");
						final_val = std::max(final_val, op_val);
						break;
					case scSMaxExpr:
						DEBUG_WITH_TYPE("loopbound", dbgs() << "Type: Nary (Max)\n");
						final_val = (unsigned) std::max((int)final_val, (int)op_val);
						break;
				}
			}
			value = final_val;
			}
			break;
		case scUDivExpr: {
			DEBUG_WITH_TYPE("loopbound", dbgs() << "Type: Div\n");
			const SCEVUDivExpr *UDivExpr = cast<SCEVUDivExpr>(equation);
			auto lhs = UDivExpr->getLHS();
			auto rhs = UDivExpr->getRHS();
			unsigned lhsval, rhsval;
			if (getSCEVBoundFromCVDomain(lhs, CVDom, lhsval) == false ||
					getSCEVBoundFromCVDomain(rhs, CVDom, rhsval) == false) {
				ret = false;
				goto end_switch;
			}
			value = lhsval / rhsval;
			}
			break;
		case scUnknown: {
			DEBUG_WITH_TYPE("loopbound", dbgs() << "Type: Unknown\n");
			const SCEVUnknown *U = cast<SCEVUnknown>(equation);
			// Try to get an argument value
			auto arg = dyn_cast<llvm::Argument>(U->getValue());
			//We only advance if we have a function argument of type int that
			//fits into a single register
			if (arg == nullptr || !arg->getType()->isIntegerTy() || arg->getType()->getPrimitiveSizeInBits() > 32) {
				AnalysisResults::getInstance().incrementResult("SCEV_unknown");
				ret = false;
				break;
			}

			ret = getValueByArgument(arg, CVDom, value);
			}
			break;
		// Cases that we don't handle
		case scTruncate:
		case scZeroExtend:
		case scSignExtend:
		case scAddRecExpr:
		case scCouldNotCompute:
		default:
			AnalysisResults::getInstance().incrementResult("SCEV_NotImplemented");
			ret = false;
	}
end_switch:
	return ret;
}

template <llvm::Triple::ArchType ISA>
std::unordered_map<Context, unsigned> LoopBoundInfoPass::getContextSensitiveBounds(
	const llvm::MachineLoop* loop,
	const llvm::SCEV* equation,
	const std::unordered_map<Context, ConstantValueDomain<ISA>>& anaInfo)
{
	std::unordered_map<Context, unsigned> ctxBounds;
	for (auto& kv : anaInfo) {
		auto ctx = kv.first;
		auto CVDomain = kv.second;
		unsigned value;
		if (getSCEVBoundFromCVDomain(equation, CVDomain, value) == true) {
			if ((unsigned)value > std::numeric_limits<int>::max()) {
				DEBUG_WITH_TYPE("loopbound", dbgs() << "The loopbound is greater than INT_MAX. Discarding\n");
			} else {
				ctxBounds.insert(std::make_pair(ctx, value));
				DEBUG_WITH_TYPE("loopbound", dbgs() << "Computed Ctx Sens Bound as: " << value << "\n");
			}
		} else {
			DEBUG_WITH_TYPE("loopbound", dbgs() << "Could not compute SCEV Bound. Incomplete Data\n");
		}
		DEBUG_WITH_TYPE("loopdump", dbgs() << "Adding loop " << *loop << ". Context: " << ctx.serialize() << "\n");
		loopContextMap[loop].insert(ctx);
	}
	return ctxBounds;
}

// Make definitions explicit here
template void LoopBoundInfoPass::computeLoopBoundFromCVDomain<Triple::ArchType::arm>(const CVAnalysisType<Triple::ArchType::arm>& cvAnaInfo);
template void LoopBoundInfoPass::computeLoopBoundFromCVDomain<Triple::ArchType::riscv32>(const CVAnalysisType<Triple::ArchType::riscv32>& cvAnaInfo);

template <llvm::Triple::ArchType ISA>
void LoopBoundInfoPass::computeLoopBoundFromCVDomain(const CVAnalysisType<ISA>& cvAnaInfo) {
	DEBUG_WITH_TYPE("loopbound", dbgs() << "+++ Computing Loop Bounds from CV Domain Now +++\n");
	AnalysisResults& ar = AnalysisResults::getInstance();
	ar.registerResult("SCEV_constant", 0);
	ar.registerResult("SCEV_argument_high", 0);
	ar.registerResult("SCEV_arg_cv", 0);
	ar.registerResult("SCEV_overflow", 0);
	ar.registerResult("SCEV_unknown", 0);
	ar.registerResult("SCEV_NotImplemented", 0);
	computeLoopBounds(upperLoopBoundsSCEV, upperLoopBoundsCtx, cvAnaInfo);
	computeLoopBounds(lowerLoopBoundsSCEV, lowerLoopBoundsCtx, cvAnaInfo);
}

template <llvm::Triple::ArchType ISA>
void LoopBoundInfoPass::computeLoopBounds(
	const std::unordered_map<const llvm::MachineLoop*,
	const llvm::SCEV*>& loopMapping,
	std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& loopBounds,
	const CVAnalysisType<ISA>& cvAnaInfo)
{
	for (auto loop : loopMapping) {
		DEBUG_WITH_TYPE("loopbound", dbgs() << "Trying to compute bounds for: " << *loop.first << "\n");
		DEBUG_WITH_TYPE("loopbound", dbgs() << "Equation is: " << *loop.second << "\n");
		auto parent_function = loop.first->getHeader()->getParent();
		std::list<MBBedge> initialedgelist;
		const MachineInstr* firstInstr = getFirstInstrInFunction(parent_function, initialedgelist);
		if (cvAnaInfo.hasAnaInfoBefore(firstInstr)) {
			DEBUG_WITH_TYPE("loopbound", dbgs() << "We have analysis info.\n");
			auto anaInfoCtx = cvAnaInfo.getAnaInfoBefore(firstInstr);
			if (!anaInfoCtx.isBottom()) {
				auto ctxBounds = getContextSensitiveBounds(loop.first, loop.second, anaInfoCtx.getAnalysisInfoPerContext());
				loopBounds.insert(std::make_pair(loop.first, ctxBounds));
			} else {
				DEBUG_WITH_TYPE("loopbound", dbgs() << "No Analysis info for bottom\n");
			}
		} else {
			DEBUG_WITH_TYPE("loopbound", dbgs() << "No Analysis info available. Will not get bounds for this loop\n");
		}
	}
}

const std::list<const llvm::MachineLoop*> LoopBoundInfoPass::getAllLoops() const {
	return maLoops;
}

bool LoopBoundInfoPass::hasLoopBoundNoCtx(const llvm::MachineLoop* loop,
		const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& loopBounds,
		const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& manualLoopBounds,
		const std::unordered_map<const llvm::MachineLoop*, unsigned>& manualLoopBoundsNoCtx) const
{

	bool hasBound = true;
	for (auto ctx : loopContextMap.at(loop)) {
		hasBound &= hasLoopBound(loop, loopBounds, manualLoopBounds, manualLoopBoundsNoCtx, ctx);
	}
	return hasBound;
}


bool LoopBoundInfoPass::hasUpperLoopBound(const llvm::MachineLoop* loop, const Context& ctx) const {
	if (ctx.isEmpty()) {
		return hasLoopBoundNoCtx(loop, upperLoopBoundsCtx, manualUpperLoopBounds, manualUpperLoopBoundsNoCtx);
	}
	return hasLoopBound(loop, upperLoopBoundsCtx, manualUpperLoopBounds, manualUpperLoopBoundsNoCtx, ctx);
}

bool LoopBoundInfoPass::hasLowerLoopBound(const llvm::MachineLoop* loop, const Context& ctx) const {
	if (ctx.isEmpty()) {
		return hasLoopBoundNoCtx(loop, lowerLoopBoundsCtx, manualLowerLoopBounds, manualLowerLoopBoundsNoCtx);
	}
	return hasLoopBound(loop, lowerLoopBoundsCtx, manualLowerLoopBounds, manualLowerLoopBoundsNoCtx, ctx);
}

bool LoopBoundInfoPass::hasLoopBound(const llvm::MachineLoop* loop,
		const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& loopBounds,
		const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& manualLoopBounds,
		const std::unordered_map<const llvm::MachineLoop*, unsigned>& manualLoopBoundsNoCtx,
		const Context& ctx) const
{
	bool hasBound = false;

	if (loopBounds.count(loop) > 0) {
		hasBound = !!loopBounds.at(loop).count(ctx);
	}

	if (!hasBound && manualLoopBounds.count(loop) > 0) {
		hasBound = !!manualLoopBounds.at(loop).count(ctx);
	}

	if (!hasBound) {
		hasBound = !!manualLoopBoundsNoCtx.count(loop);
	}

	return hasBound;
}

unsigned LoopBoundInfoPass::getLoopBoundNoCtx(const llvm::MachineLoop* loop,
		const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& loopBounds,
		const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& manualLoopBounds,
		const std::unordered_map<const llvm::MachineLoop*, unsigned>& manualLoopBoundsNoCtx) const
{
	unsigned maxBound = 0;
	for (auto ctx : loopContextMap.at(loop)) {
		maxBound = std::max(maxBound, getLoopBound(loop, loopBounds, manualLoopBounds, manualLoopBoundsNoCtx, ctx));
	}
	return maxBound;
}

unsigned LoopBoundInfoPass::getUpperLoopBound(const llvm::MachineLoop* loop, const Context& ctx) const {
	if (ctx.isEmpty()) {
		return getLoopBoundNoCtx(loop, upperLoopBoundsCtx, manualUpperLoopBounds, manualUpperLoopBoundsNoCtx);
	}
	return getLoopBound(loop, upperLoopBoundsCtx, manualUpperLoopBounds, manualUpperLoopBoundsNoCtx, ctx);
}

unsigned LoopBoundInfoPass::getLowerLoopBound(const llvm::MachineLoop* loop, const Context &ctx) const {
	if (ctx.isEmpty()) {
		return getLoopBoundNoCtx(loop, lowerLoopBoundsCtx, manualLowerLoopBounds, manualLowerLoopBoundsNoCtx);
	}
	return getLoopBound(loop, lowerLoopBoundsCtx, manualLowerLoopBounds, manualLowerLoopBoundsNoCtx, ctx);
}

unsigned LoopBoundInfoPass::getLoopBound(const llvm::MachineLoop* loop,
		const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& loopBounds,
		const std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& manualLoopBounds,
		const std::unordered_map<const llvm::MachineLoop*, unsigned>& manualLoopBoundsNoCtx,
		const Context& ctx) const {
	assert (hasLoopBound(loop, loopBounds, manualLoopBounds, manualLoopBoundsNoCtx, ctx) && "We do not have a loop bound, check before");
	assert (!ctx.isEmpty() && "Don't pass an empty context to getLoopBound");
	unsigned bound;
	bool foundBoundManual = false;

	if (manualLoopBoundsNoCtx.count(loop) > 0) {
		bound = manualLoopBoundsNoCtx.at(loop);
		foundBoundManual = true;
	}

	if (manualLoopBounds.count(loop) > 0) {
		auto manualBounds = manualLoopBounds.at(loop);
		if (manualBounds.count(ctx) > 0) {
			auto manualBoundVal = manualBounds.at(ctx);
			if (foundBoundManual && (bound != manualBoundVal)) {
				errs() << "Warning: Both CtxSens and Plain manual loop bounds were found and bounds differ! (CtxSens was used) for:\n"
				       << *loop << "\n";
			}
			bound = manualBoundVal;
		}
	}


	if (loopBounds.count(loop) > 0) {
		auto ctxBounds = loopBounds.at(loop);
		if (ctxBounds.count(ctx) > 0) {
			auto autoBound = ctxBounds.at(ctx);
			if (foundBoundManual && (autoBound != bound)) {
				errs() << "Warnings Both automatic and manual loop bounds were found and bounds differ! (Automatic used) for:\n"
					   << loop->getHeader()->getParent()->getName().str() << " | " << *loop << "\n";
			}
			bound = autoBound;
		}
	}

	return bound;
}

void LoopBoundInfoPass::dump(std::ostream& mystream) const {
	raw_os_ostream llvmstream(mystream);
	llvmstream << "We found the following loop bounds automatically:\n";
	std::set<std::string> orderedLoopBoundOutput;
	for (auto loopMap : loopContextMap) {
		auto func = loopMap.first->getHeader()->getParent();
		for (auto ctx : loopMap.second) {
			if (hasUpperLoopBound(loopMap.first, ctx)) {
				orderedLoopBoundOutput.insert(
						"# In function " + func->getName().str()
					+ ", loop:\n  " + getLoopDesc(loopMap.first)
					+ " with context " + ctx.serialize()
					+ "\nthe loop header is executed at most " + std::to_string(getUpperLoopBound(loopMap.first, ctx)) + " times.\n"
				);
			} else {
				orderedLoopBoundOutput.insert(
						"# In function " + func->getName().str()
					+ ", loop:\n  " + getLoopDesc(loopMap.first)
					+ " with context " + ctx.serialize()
					+ "\nthe loop header execution count cannot be bounded.\n"
				);
			}
		}
	}
	for (auto& output : orderedLoopBoundOutput) {
		llvmstream << output;
	}
}

std::string LoopBoundInfoPass::getLoopDesc(const llvm::MachineLoop* loop) const{
	std::string res;
	auto loopBlocks = loop->getBlocks();
	for (auto &block : loopBlocks){
		res += "BB#";
		res += std::to_string(block->getNumber());
		res += ":";
		res += block->getName();
		if (block == loop->getHeader()){
			res += "<header>";
		}
		if (block == loop->getLoopLatch()){
			res += "<latch>";
		}
		if (loop->isLoopExiting(block)){
			res += "<exiting>";
		}
		res += "_";
	}	
	return res.substr(0, res.size()-1);
}

void LoopBoundInfoPass::dumpNonUpperBoundLoops(std::ostream& mystream, std::ostream& mystream2) const {
	dumpNonBoundLoops(mystream, mystream2, &TimingAnalysisPass::LoopBoundInfoPass::hasUpperLoopBound, &TimingAnalysisPass::LoopBoundInfoPass::getUpperLoopBound);
}

void LoopBoundInfoPass::dumpNonLowerBoundLoops(std::ostream& mystream, std::ostream& mystream2) const {
	dumpNonBoundLoops(mystream, mystream2, &TimingAnalysisPass::LoopBoundInfoPass::hasLowerLoopBound, &TimingAnalysisPass::LoopBoundInfoPass::getLowerLoopBound);
}

void LoopBoundInfoPass::dumpNonBoundLoops(std::ostream& mystream, std::ostream& mystream2, bool (LoopBoundInfoPass::* hasLoopBoundFkt) (const llvm::MachineLoop* loop, const Context& ctx) const, unsigned (LoopBoundInfoPass::* getLoopBoundFkt) (const llvm::MachineLoop* loop, const Context& ctx) const ) const
{
	DEBUG_WITH_TYPE("loopdump", dbgs() << "Beginning loop dump\n");
	CallGraph& cg = CallGraph::getGraph();
	raw_os_ostream llvmstream(mystream);
	raw_os_ostream llvmstream2(mystream2);
	llvmstream << "# Type: ContextSensitive\n";
	llvmstream2 << "# Type: Normal\n";
	for (auto loopMap : loopMapping) {
		std::string loopInfoString;
		raw_string_ostream loopInfo(loopInfoString);
		auto header = loopMap.first->getHeader();
		if (!cg.reachableFromEntryPoint(header->getParent())) {
			continue;
		}
		DEBUG_WITH_TYPE("loopdump", dbgs() << "[Loop]: " << *loopMap.first);
		loopInfo << header->getParent()->getName() << "|";
		loopInfo << "Loop ";
		DEBUG_WITH_TYPE("loopdump", dbgs() << "[LFunc]: " <<  header->getParent()->getName() << "\n");
		if (!header->empty()
				&& !header->front().isTransient()
				&& !(header->front().mayLoad() || header->front().mayStore())
				&& header->front().getDebugLoc()
				&& header->front().getDebugLoc().getLine() != 0)
		{
			auto dbgLoc = header->front().getDebugLoc();
			std::string filename = getFilenameFromDebugLoc(dbgLoc);
			loopInfo << "in file " << filename;
			loopInfo << " near line " << dbgLoc.getLine() << "|";
		} else {
			bool unknown = true;
			for (auto& currInstr : *header) {
				if (currInstr.isTransient() || currInstr.mayLoad() || currInstr.mayStore()) {
					continue;
				}
				if (currInstr.getDebugLoc() && currInstr.getDebugLoc().getLine() != 0) {
					std::string filename = getFilenameFromDebugLoc(currInstr.getDebugLoc());
					loopInfo << "in file " << filename;
					loopInfo << " near line " << currInstr.getDebugLoc().getLine() << "|";
					unknown = false;
					break;
				}
			}
			if (unknown) {
				// Try to get information from the middle-end loops
				if (loopMapping.count(loopMap.first) > 0) {
					std::cout << "Trying to get info from middle-end loops\n";
					auto irloop = loopMapping.at(loopMap.first);
					auto irinstr = irloop->getHeader()->getTerminator();
					if (irinstr != nullptr && irinstr->getDebugLoc() && irinstr->getDebugLoc().getLine() != 0) {
						std::string filename = getFilenameFromDebugLoc(irinstr->getDebugLoc());
						loopInfo << "in file " << filename;
						loopInfo << " near line " << irinstr->getDebugLoc().getLine() << "|";
						unknown = false;
					} else {
						assert(irinstr != nullptr && "IR Instr is nullptr");
						assert(irinstr->getDebugLoc() && "Could not get DebugLoc from IR");
						assert(false);
					}
				}
				// If still unknown
				if (unknown) {
					loopInfo << "in file UNKNOWN near line 0|";
				}
			}
		}
		DEBUG_WITH_TYPE("loopdump", dbgs() << "[LSTR]: " << loopInfo.str() << "\n");

		loopInfo << getLoopDesc(loopMap.first) << "|";
		auto maxBound = -1;

		if (loopContextMap.count(loopMap.first) > 0) {
			assert(loopContextMap.at(loopMap.first).size() > 0);
			// Format is: "Function|LoopDescription|LoopContext|Bound"\n
			for (auto ctx : loopContextMap.at(loopMap.first)) {
				DEBUG_WITH_TYPE("loopdump", dbgs() << "Context: " << ctx.serialize() << "\n");
				llvmstream << loopInfo.str() << ctx.serialize() << "|";
				if ((this->*hasLoopBoundFkt)(loopMap.first, ctx)){
					auto bound = (this->*getLoopBoundFkt)(loopMap.first, ctx);
					llvmstream << bound;
					maxBound = std::max(maxBound, (int)bound);
				} else {
					llvmstream << "-1";
				}
				llvmstream << "\n";
			}
		}
		DEBUG_WITH_TYPE("loopdump", dbgs() << "[LCTX]: " << loopInfo.str() << maxBound << "\n");
		llvmstream2 << loopInfo.str() << maxBound << "\n";
		/* DEBUG_WITH_TYPE("loopdump", dbgs() << llvmstream2.str() << "\n"); */
	}
}

void LoopBoundInfoPass::parseManualUpperLoopBounds(const char *filename) {
	DEBUG_WITH_TYPE("loopbound", dbgs() << "# Parse upper loop bounds\n");
	parseManualLoopBounds(filename, manualUpperLoopBounds, manualUpperLoopBoundsNoCtx);
}

void LoopBoundInfoPass::parseManualLowerLoopBounds(const char *filename) {
	DEBUG_WITH_TYPE("loopbound", dbgs() << "# Parse lower loop bounds\n");
	parseManualLoopBounds(filename, manualLowerLoopBounds, manualLowerLoopBoundsNoCtx);
}

void LoopBoundInfoPass::parseNormalManualLoopBounds(std::ifstream& file,
		std::unordered_map<const llvm::MachineLoop*, unsigned>& manualLoopBounds)
{
	DEBUG_WITH_TYPE("loopbound", dbgs() << "## Parsing Normal (Non CtxSens) Manual Bounds\n");

	typedef boost::tokenizer< boost::escaped_list_separator<char> > Tokenizer;
	boost::escaped_list_separator<char> sep('\\','|','"');

	for(std::string line; getline(file, line);) {
		DEBUG_WITH_TYPE("loopbound", dbgs() << "# Processing line: " << line << "\n");
		std::vector<std::string> vect;
		Tokenizer token(line, sep);
		vect.assign(token.begin(), token.end());
		assert (vect.size() == 4 && "CSV file was corrupted!");
		std::string functionName = vect[0];
		// vect[1] ist only debug location printing
		std::string loopDesc = vect[2];
		int value = std::stoi(vect[3]);
		// -2 corresponds to do not care about this loop
		if (value == -2) {
			continue;
		}
		// Other negative values correspond to undefined
		assert (value >= 0 && "[Error] Loop bound is not properly defined.\n");
		// Add non-negative manual loop bound
		unsigned bound = (unsigned) value;
		for (auto loop : maLoops) {
			DEBUG_WITH_TYPE("loopbound", dbgs() << "# Checking equivalence for loop: " << *loop);
			if (loop->getHeader()->getParent()->getName() == functionName){
				DEBUG_WITH_TYPE("loopbound", dbgs() << "# Matching function name\n");
				DEBUG_WITH_TYPE("loopbound", dbgs() << "# Checking loop descriptions:\n" << loopDesc << "\n"
				                                    << getLoopDesc(loop) << "\n";);
				if (getLoopDesc(loop) == loopDesc) {
					DEBUG_WITH_TYPE("loopbound", dbgs() << "# Matching loop description\n");
					const MachineLoop *maloop = loop;
					manualLoopBounds.insert(std::make_pair(maloop, bound));
					DEBUG_WITH_TYPE("loopbound", dbgs() << "Add bound for " << maloop << "(" << *maloop << ")" << " as " << bound << "\n" );
					goto end_loop;
				}
			}
		}
end_loop:
		((void)0);
	}

}

void LoopBoundInfoPass::parseCtxSensManualLoopBounds(std::ifstream& file,
		std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& manualLoopBounds)
{
	DEBUG_WITH_TYPE("loopbound", dbgs() << "## Parsing Context Sensitive Manual Bounds\n");

	typedef boost::tokenizer< boost::escaped_list_separator<char> > Tokenizer;
	boost::escaped_list_separator<char> sep('\\','|','"');

	for(std::string line; getline(file, line);) {
		DEBUG_WITH_TYPE("loopbound", dbgs() << "# Processing line: " << line << "\n");
		std::vector<std::string> vect;
		Tokenizer token(line, sep);
		vect.assign(token.begin(), token.end());
		assert (vect.size() == 5 && "CSV file was corrupted!");
		std::string functionName = vect[0];
		// vect[1] ist only debug location printing
		std::string loopDesc = vect[2];
		std::string context = vect[3];
		int value = std::stoi(vect[4]);
		// -2 corresponds to do not care about this loop
		if (value == -2) {
			continue;
		}
		// Other negative values correspond to undefined
		assert (value >= 0 && "[Error] Loop bound is not properly defined.\n");
		// Add non-negative manual loop bound
		unsigned bound = (unsigned) value;
		for (auto loop : maLoops){
			DEBUG_WITH_TYPE("loopbound", dbgs() << "# Checking equivalence for loop: " << *loop);
			if (loop->getHeader()->getParent()->getName() == functionName){
				DEBUG_WITH_TYPE("loopbound", dbgs() << "# Matching function name\n");
				DEBUG_WITH_TYPE("loopbound", dbgs() << "# Checking loop descriptions:\n" << loopDesc << "\n"
				                                    << getLoopDesc(loop) << "\n";);
				if (getLoopDesc(loop) == loopDesc){
					DEBUG_WITH_TYPE("loopbound", dbgs() << "# Matching loop description\n");
					DEBUG_WITH_TYPE("loopbound", dbgs() << "# Checking loop contexts:\n");
					for (auto ctx : loopContextMap.at(loop)) {
						DEBUG_WITH_TYPE("loopbound", dbgs() << "# Checking context: " << ctx.serialize() << "\n");
						if (ctx.serialize() == context) {
							DEBUG_WITH_TYPE("loopbound", dbgs() << "# Matching context found\n");
							const MachineLoop *maloop = loop;
							manualLoopBounds[maloop].insert(std::make_pair(ctx, bound));
							DEBUG_WITH_TYPE("loopbound", dbgs() << "Add bound for " << maloop << "(" << *maloop << ") in context " << context << " as " << bound << "\n" );
							goto end_loop;
						}
					}
				}
			}
		}
end_loop:
		((void)0);
	}

}

void LoopBoundInfoPass::parseManualLoopBounds(const char *filename,
		std::unordered_map<const llvm::MachineLoop*, std::unordered_map<Context, unsigned>>& manualLoopBounds,
		std::unordered_map<const llvm::MachineLoop*, unsigned>& manualLoopBoundsNoCtx)
{
	DEBUG_WITH_TYPE("loopbound", dbgs() << "# Parse bounds\n");

	std::ifstream file(filename);
	if (!(file.good())){
		errs() << "File could not be opened!\n";
		return;
	}
	std::string line;
	getline(file, line);

	if (line == "# Type: ContextSensitive") {
		LoopBoundInfoPass::parseCtxSensManualLoopBounds(file, manualLoopBounds);
	} else if (line.compare("# Type: Normal") == 0) {
		LoopBoundInfoPass::parseNormalManualLoopBounds(file, manualLoopBoundsNoCtx);
	} else {
		assert(false && "Manual Loopbounds should define type in (ContextSensitive, Normal)");
	}
	file.close();
}
}

FunctionPass* llvm::createLoopBoundInfoPass() {
	TimingAnalysisPass::loopBoundInfoPass = new TimingAnalysisPass::LoopBoundInfoPass();
	return TimingAnalysisPass::loopBoundInfoPass;
}
